use super::Storage;
use crate::{
    payment::PaymentTo,
    storage::{Batch, Error, ReceivedPaymentStatus},
    AccountId, BaseTx, PaymentId, RevId, Revision, Tag, TxId,
};
use std::{collections::HashMap, marker::PhantomData, sync::Arc};

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum Ids {
    Transaction(TxId),
    Payment(PaymentId),
    Account(AccountId),
}

pub struct CacheBatch<'a, B>
where
    B: Batch<'a> + Send,
{
    inner: B,
    storage: Arc<Storage>,
    to_invalidate: HashMap<Ids, ()>,
    _phantom: PhantomData<&'a ()>,
}

impl<'a, B> CacheBatch<'a, B>
where
    B: Batch<'a> + Send,
{
    pub fn new(batch: B, storage: Arc<Storage>) -> Self {
        Self {
            inner: batch,
            storage,
            to_invalidate: HashMap::new(),
            _phantom: PhantomData,
        }
    }
}

#[async_trait::async_trait]
impl<'a, B> Batch<'a> for CacheBatch<'a, B>
where
    B: Batch<'a> + Send,
{
    async fn commit(self) -> Result<(), Error> {
        let mut payments = self.storage.payments.write().await;
        let mut balances = self.storage.balances.write().await;
        let mut revision_by_transactions = self.storage.revisions_by_transactions.write().await;

        self.inner.commit().await?;

        for (id, _) in self.to_invalidate {
            match id {
                Ids::Transaction(id) => {
                    revision_by_transactions.remove(&id);
                }
                Ids::Payment(id) => {
                    payments.remove(&id);
                }
                Ids::Account(id) => {
                    balances.remove(&id);
                }
            }
        }

        Ok(())
    }

    async fn relate_account_to_transaction(
        &mut self,
        transaction_id: &TxId,
        account: &AccountId,
        typ: crate::Type,
    ) -> Result<(), Error> {
        self.to_invalidate
            .insert(Ids::Transaction(transaction_id.clone()), ());
        self.to_invalidate.insert(Ids::Account(account.clone()), ());
        self.inner
            .relate_account_to_transaction(transaction_id, account, typ)
            .await
    }

    async fn rollback(self) -> Result<(), Error> {
        self.inner.rollback().await
    }

    async fn update_transaction_payments(
        &mut self,
        transaction_id: &TxId,
        create_status: ReceivedPaymentStatus,
        spend_status: ReceivedPaymentStatus,
    ) -> Result<(usize, usize), Error> {
        self.to_invalidate
            .insert(Ids::Transaction(transaction_id.clone()), ());
        self.inner
            .update_transaction_payments(transaction_id, create_status, spend_status)
            .await
    }

    async fn get_payment_status(
        &mut self,
        transaction_id: &TxId,
    ) -> Result<Option<ReceivedPaymentStatus>, Error> {
        self.to_invalidate
            .insert(Ids::Transaction(transaction_id.clone()), ());
        self.inner.get_payment_status(transaction_id).await
    }

    async fn spend_payments(
        &mut self,
        transaction_id: &TxId,
        spends: Vec<PaymentId>,
        status: ReceivedPaymentStatus,
    ) -> Result<(), Error> {
        for spend in &spends {
            self.to_invalidate.insert(Ids::Payment(spend.clone()), ());
        }
        self.inner
            .spend_payments(transaction_id, spends, status)
            .await
    }

    async fn create_payments(
        &mut self,
        transaction_id: &TxId,
        recipients: &[PaymentTo],
        status: ReceivedPaymentStatus,
    ) -> Result<(), Error> {
        for recipient in recipients {
            self.to_invalidate
                .insert(Ids::Account(recipient.to.clone()), ());
        }
        self.inner
            .create_payments(transaction_id, recipients, status)
            .await
    }

    async fn store_base_transaction(
        &mut self,
        transaction_id: &TxId,
        transaction: &BaseTx,
    ) -> Result<(), Error> {
        self.to_invalidate
            .insert(Ids::Transaction(transaction_id.clone()), ());
        self.inner
            .store_base_transaction(transaction_id, transaction)
            .await
    }

    async fn store_revision(
        &mut self,
        revision_id: &RevId,
        revision: &Revision,
    ) -> Result<(), Error> {
        self.to_invalidate
            .insert(Ids::Transaction(revision.transaction_id.clone()), ());
        self.inner.store_revision(revision_id, revision).await
    }

    async fn update_transaction_revision(
        &mut self,
        transaction_id: &TxId,
        revision_id: &RevId,
        previous_rev_id: Option<&RevId>,
    ) -> Result<(), Error> {
        self.to_invalidate
            .insert(Ids::Transaction(transaction_id.clone()), ());
        self.inner
            .update_transaction_revision(transaction_id, revision_id, previous_rev_id)
            .await
    }

    async fn tag_transaction(
        &mut self,
        transaction_id: &TxId,
        base_tx: &BaseTx,
        tags: &[Tag],
    ) -> Result<(), Error> {
        self.to_invalidate
            .insert(Ids::Transaction(transaction_id.clone()), ());
        self.inner
            .tag_transaction(transaction_id, base_tx, tags)
            .await
    }
}
