//! Cache storage implementation.
use crate::{
    amount::AmountCents,
    storage::{self, Error},
    AccountId, Amount, Asset, BaseTx, Filter, PaymentFrom, PaymentId, RevId, Revision, TxId,
};
use std::{collections::HashMap, sync::Arc};
use tokio::sync::RwLock;

mod batch;

type SharedHashMap<K, V> = RwLock<HashMap<K, V>>;

/// Cache storage implementation.
///
/// This cache will wrap any actual storage and will add an in memory caching to
/// offer better performance. This layer will also manager cache eviction and
/// invalidation.
pub struct Cache<S>
where
    S: storage::Storage + Sync + Send,
{
    inner: S,
    storage: Arc<Storage>,
}

#[derive(Debug, Default)]
/// Cache storage.
pub struct Storage {
    payments: SharedHashMap<PaymentId, PaymentFrom>,
    balances: SharedHashMap<AccountId, Vec<Amount>>,
    revisions_by_transactions: SharedHashMap<TxId, Vec<RevId>>,
}

impl<S> Cache<S>
where
    S: storage::Storage + Sync + Send,
{
    /// Create a new cache storage.
    pub fn new(storage: S) -> Self {
        Self {
            storage: Arc::new(Storage::default()),
            inner: storage,
        }
    }
}

#[async_trait::async_trait]
impl<S> storage::Storage for Cache<S>
where
    S: storage::Storage + Sync + Send,
{
    type Batch<'a> = batch::CacheBatch<'a, S::Batch<'a>>
    where
        Self: 'a;

    async fn begin<'a>(&'a self) -> Result<Self::Batch<'a>, Error> {
        Ok(batch::CacheBatch::new(
            self.inner.begin().await?,
            self.storage.clone(),
        ))
    }

    async fn get_balance(&self, account_id: &AccountId) -> Result<Vec<Amount>, Error> {
        let cache = self.storage.balances.read().await;
        if let Some(balances) = cache.get(account_id).cloned() {
            Ok(balances)
        } else {
            drop(cache);
            let result = self.inner.get_balance(account_id).await?;
            self.storage
                .balances
                .write()
                .await
                .insert(account_id.clone(), result.clone());
            Ok(result)
        }
    }

    async fn get_revisions(&self, transaction_id: &TxId) -> Result<Vec<RevId>, Error> {
        let cache = self.storage.revisions_by_transactions.read().await;
        if let Some(revisions) = cache.get(transaction_id) {
            Ok(revisions.clone())
        } else {
            drop(cache);
            let revisions = self.inner.get_revisions(transaction_id).await?;
            self.storage
                .revisions_by_transactions
                .write()
                .await
                .insert(transaction_id.clone(), revisions.clone());
            Ok(revisions)
        }
    }

    async fn find_base_tx_and_revision(
        &self,
        filter: Filter,
    ) -> Result<Vec<(BaseTx, Revision)>, Error> {
        self.inner.find_base_tx_and_revision(filter).await
    }

    async fn get_negative_unspent_payments(
        &self,
        account: &AccountId,
        asset: &Asset,
    ) -> Result<Vec<PaymentFrom>, Error> {
        self.inner
            .get_negative_unspent_payments(account, asset)
            .await
    }

    async fn get_positive_unspent_payments(
        &self,
        account: &AccountId,
        asset: &Asset,
        target_amount: AmountCents,
    ) -> Result<Vec<PaymentFrom>, Error> {
        self.inner
            .get_positive_unspent_payments(account, asset, target_amount)
            .await
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::{storage::sqlite::SQLite, storage_test_suite, tests::deposit, Ledger};
    use sqlx::sqlite::{SqliteConnectOptions, SqlitePoolOptions};
    use std::fs::remove_file;

    async fn get_instance(test: &str) -> Cache<SQLite> {
        let path = format!("/tmp/cache-{}.db", test);
        let _ = remove_file(&path);
        let settings = path
            .parse::<SqliteConnectOptions>()
            .expect("valid settings")
            .create_if_missing(true);

        let pool = SqlitePoolOptions::new()
            .connect_with(settings)
            .await
            .expect("pool");

        let db = SQLite::new(pool);
        db.setup().await.expect("setup");
        Cache::new(db)
    }

    storage_test_suite!();

    pub async fn get_ledger_and_asset_manager() -> Arc<Ledger<Cache<SQLite>>> {
        let pool = SqlitePoolOptions::new()
            .max_connections(1)
            .idle_timeout(None)
            .max_lifetime(None)
            .connect(":memory:")
            .await
            .expect("pool");

        let db = SQLite::new(pool);
        db.setup().await.expect("setup");
        Ledger::new(Cache::new(db).into())
    }

    #[tokio::test]
    async fn balances_updates() {
        let source = "account1".parse::<AccountId>().expect("account");
        let dest = "account2".parse::<AccountId>().expect("account");
        let fee = "fee".parse::<AccountId>().expect("account");
        let ledger = get_ledger_and_asset_manager().await;
        let usd: Asset = "USD/2".parse().expect("asset");

        deposit(&ledger, &source, usd.from_human("10").expect("amount")).await;
        deposit(&ledger, &source, usd.from_human("20").expect("amount")).await;

        assert_eq!(
            vec![usd.from_human("30").expect("amount"),],
            ledger.get_balance(&source).await.expect("balance")
        );

        ledger
            .new_transaction(
                "Exchange one".to_owned(),
                "settled".into(),
                vec![(source.clone(), usd.from_human("13").expect("amount"))],
                vec![
                    (dest.clone(), usd.from_human("12.5").expect("amount")),
                    (fee.clone(), usd.from_human("0.5").expect("amount")),
                ],
            )
            .await
            .expect("valid tx");

        assert_eq!(
            vec![usd.from_human("17").expect("amount"),],
            ledger.get_balance(&source).await.expect("balance")
        );
        assert_eq!(
            vec![usd.from_human("12.5").expect("amount"),],
            ledger.get_balance(&dest).await.expect("balance")
        );
        assert_eq!(
            vec![usd.from_human("0.5").expect("amount"),],
            ledger.get_balance(&fee).await.expect("balance")
        );
    }
}
