//! Storage layer trait
use crate::{
    amount::AmountCents, payment::PaymentTo, transaction::Type, AccountId, Amount, Asset, BaseTx,
    Filter, PaymentFrom, PaymentId, RevId, Revision, Tag, Transaction, TxId,
};
//use chrono::{DateTime, Utc};
use serde::Serialize;

pub mod cache;
mod cursor;

#[cfg(any(feature = "sqlite", test))]
pub mod sqlite;
pub use self::cache::Cache;
pub use self::cursor::Cursor;
#[cfg(any(feature = "sqlite", test))]
pub use self::sqlite::SQLite;

#[derive(Clone, Copy, PartialEq, Eq, Debug, Serialize)]
/// Storage Payment Status
pub enum ReceivedPaymentStatus {
    ///  The payment is available to be spent
    Spendable,
    /// The payment is not spentable and will not be part of the balance
    Failed,
    /// The payment is unspendable temporarily, it can be updated to any other status
    Locked,
    /// The transaction has been spent
    Spent,
}

/// Serializes an object to bytes using the default serialization method
pub fn to_bytes<T>(val: &T) -> Result<Vec<u8>, Error>
where
    T: serde::Serialize + borsh::BorshSerialize,
{
    borsh::to_vec(val).map_err(|e| Error::Encoding(e.to_string()))
}

/// Deserializes an object from bytes using the default deserialization method
pub fn from_bytes<T>(val: &[u8]) -> Result<T, Error>
where
    T: serde::de::DeserializeOwned + borsh::BorshDeserialize,
{
    #[allow(clippy::useless_asref)]
    borsh::BorshDeserialize::deserialize(&mut val.as_ref())
        .map_err(|e| Error::Encoding(e.to_string()))
}

impl From<ReceivedPaymentStatus> for u32 {
    fn from(val: ReceivedPaymentStatus) -> Self {
        match val {
            ReceivedPaymentStatus::Spendable => 0,
            ReceivedPaymentStatus::Locked => 20,
            ReceivedPaymentStatus::Spent => 30,
            ReceivedPaymentStatus::Failed => 40,
        }
    }
}

impl TryFrom<u32> for ReceivedPaymentStatus {
    type Error = Error;

    fn try_from(value: u32) -> Result<Self, Self::Error> {
        match value {
            0 => Ok(ReceivedPaymentStatus::Spendable),
            20 => Ok(ReceivedPaymentStatus::Locked),
            30 => Ok(ReceivedPaymentStatus::Spent),
            40 => Ok(ReceivedPaymentStatus::Failed),
            _ => Err(Error::Storage(format!("Invalid status: {}", value))),
        }
    }
}

#[derive(thiserror::Error, Debug, Serialize)]
/// Storage error
pub enum Error {
    #[error("Storage error: {0}")]
    /// Any error related to the storage layer, with string
    Storage(String),

    #[error("I/O error: {0}")]
    #[serde(serialize_with = "crate::serialize_error_to_string")]
    /// Any error related to the storage layer, with std::io::Error
    Io(#[from] std::io::Error),

    #[error("Invalid date format: {0}")]
    /// The date format is invalid
    InvalidDate(String),

    #[error("Invalid id: {0}")]
    #[serde(serialize_with = "crate::serialize_error_to_string")]
    /// Invalid Id
    InvalidId(#[from] crate::id::Error),

    #[error("Spend payment: {0}")]
    /// The requested payment has been spent already
    SpendPayment(String),

    #[error("No storage update when expecting")]
    /// No update was performed when expecting one
    NoUpdate,

    #[error("Math error, overflow")]
    /// Overflow error
    Overflow,

    #[error("Math error, underflow")]
    /// Underflow error
    Underflow,

    #[error("Record not found")]
    /// The requested record was not found
    NotFound,

    #[error("Internal error with encoding: {0}")]
    /// Error with the encoding/decoding
    Encoding(String),

    #[error("Not enough unspent payments (missing {0} cents)")]
    /// TODO: Convert the AmountCents to Amount for better error reporting upstream
    NotEnoughUnspentPayments(AmountCents),
}

#[async_trait::async_trait]
/// Batches
///
/// Batches are a group of updates to the databases, in which all of the are
/// executed or none. Same concept as a rdbms transaction.
pub trait Batch<'a> {
    /// The current batch is discarded
    async fn rollback(self) -> Result<(), Error>;

    /// The batch is commited into the storage layer and it is consumed.
    async fn commit(self) -> Result<(), Error>;

    /// Flag the given payments as spent by the given transaction.
    async fn spend_payments(
        &mut self,
        transaction_id: &TxId,
        spends: Vec<PaymentId>,
        status: ReceivedPaymentStatus,
    ) -> Result<(), Error>;

    /// Create a new list of payments
    async fn create_payments(
        &mut self,
        transaction_id: &TxId,
        recipients: &[PaymentTo],
        status: ReceivedPaymentStatus,
    ) -> Result<(), Error>;

    /// Updates all payments related to a transaction
    ///
    /// This function will update the created and spent payments related to a given transaction.
    ///
    /// This function must check the previous status before changing it. For instance, Failed and
    /// Spent are final statuses, making the transaction read-only, going forward. The status
    /// from NotSpentable to Spendable can only be done by the parent transaction, and not other.
    async fn update_transaction_payments(
        &mut self,
        transaction_id: &TxId,
        create_status: ReceivedPaymentStatus,
        spend_status: ReceivedPaymentStatus,
    ) -> Result<(usize, usize), Error>;

    /// Returns the stats of a payment from the point of view of the on-going transaction
    async fn get_payment_status(
        &mut self,
        transaction_id: &TxId,
    ) -> Result<Option<ReceivedPaymentStatus>, Error>;

    /// Stores a transaction
    async fn store_base_transaction(
        &mut self,
        transaction_id: &TxId,
        transaction: &BaseTx,
    ) -> Result<(), Error>;

    /// Persists a revision
    async fn store_revision(
        &mut self,
        revision_id: &RevId,
        revision: &Revision,
    ) -> Result<(), Error>;

    /// Updates the revision of a transaction. Both the revision and the transaction must exist.
    async fn update_transaction_revision(
        &mut self,
        transaction_id: &TxId,
        revision_id: &RevId,
        previous_rev_id: Option<&RevId>,
    ) -> Result<(), Error>;

    /// Sets the tags for a given transaction. Any tag not included in this
    /// vector should be removed
    async fn tag_transaction(
        &mut self,
        transaction_id: &TxId,
        base_tx: &BaseTx,
        tags: &[Tag],
    ) -> Result<(), Error>;

    /// Creates a relationship between an account and a transaction.
    ///
    /// This is a chance to build an index which relates accounts to
    /// transactions (spending or receiving accounts). This information is used
    /// to query ledger information per account and to enhance any future cache
    /// layer.
    async fn relate_account_to_transaction(
        &mut self,
        transaction: &TxId,
        account: &AccountId,
        typ: Type,
    ) -> Result<(), Error>;
}

#[async_trait::async_trait]
/// Main storage layer
pub trait Storage {
    /// The batch trait.
    ///
    /// This batch is needed to perform the changes in the transactions and
    /// payments in a atomic way
    type Batch<'a>: Batch<'a> + Send
    where
        Self: 'a;

    /// Begins a transaction
    ///
    /// A transaction prepares the storage for a batch of operations, where all
    /// the operations are persisted or none.
    ///
    /// The transaction is returned as a Batch object, which is used to perform
    /// the changes in the transactions and payments.
    ///
    /// The batch has also a rollback
    async fn begin<'a>(&'a self) -> Result<Self::Batch<'a>, Error>;

    /// Returns the balances for a given account
    async fn get_balance(&self, account: &AccountId) -> Result<Vec<Amount>, Error>;

    /// Returns all the negative payments that are unspent by any account
    ///
    /// If any unspent negative deposit exists with the given account and asset, it must be spend in
    /// the transaction.
    async fn get_negative_unspent_payments(
        &self,
        account: &AccountId,
        asset: &Asset,
    ) -> Result<Vec<PaymentFrom>, Error>;

    /// Returns a list of unspent payments for a given account and asset.
    ///
    /// The payments should be returned sorted by ascending amount, this bit is
    /// important to make sure that all negative payments are always taken into
    /// account. It will also improve the database by using many small payments
    /// instead of a few large ones, which will make the database faster leaving
    /// fewer unspent payments when checking for balances.
    async fn get_positive_unspent_payments(
        &self,
        account: &AccountId,
        asset: &Asset,
        target_amount: AmountCents,
    ) -> Result<Vec<PaymentFrom>, Error>;

    /// Returns a list of unspent payments for a given account and asset.
    ///
    /// This list includes all the negative unspent payments and the list of positive unspent
    /// payments to cover the target_amount
    async fn get_unspent_payments(
        &self,
        account: &AccountId,
        asset: &Asset,
        target_amount: AmountCents,
    ) -> Result<Vec<PaymentFrom>, Error> {
        let mut payments = self.get_negative_unspent_payments(account, asset).await?;
        let target_amount = target_amount
            .checked_add(
                payments
                    .iter()
                    .map(|payment| payment.amount.cents())
                    .sum::<AmountCents>()
                    .checked_abs()
                    .ok_or(Error::Overflow)?,
            )
            .ok_or(Error::Overflow)?;

        payments.extend(
            self.get_positive_unspent_payments(account, asset, target_amount)
                .await?,
        );
        Ok(payments)
    }

    /// Find transactions with a given filter
    async fn find_base_tx_and_revision(
        &self,
        mut filter: Filter,
    ) -> Result<Vec<(BaseTx, Revision)>, Error>;

    /// List all revisions for a given transaction
    async fn get_revisions(&self, transaction_id: &TxId) -> Result<Vec<RevId>, Error>;

    /// Returns a revision with a transaction object by id
    async fn find(&self, filter: Filter) -> Result<Vec<Transaction>, Error> {
        let mut results = vec![];
        for (base_tx, revision) in self.find_base_tx_and_revision(filter).await?.into_iter() {
            results.push(
                (
                    self.get_revisions(&revision.transaction_id).await?,
                    base_tx,
                    revision,
                )
                    .try_into()
                    .map_err(|e: crate::transaction::Error| Error::Encoding(e.to_string()))?,
            );
        }

        Ok(results)
    }
}

#[cfg(test)]
pub mod test {
    use super::*;
    use crate::{config::Config, status::StatusManager, Ledger, Transaction};
    use rand::Rng;
    use std::{collections::HashMap, time::Duration};
    use tokio::time::sleep;

    #[macro_export]
    macro_rules! storage_unit_test {
        ($name:ident) => {
            #[tokio::test]
            async fn $name() {
                let ledger = get_instance(stringify!($name)).await;
                $crate::storage::test::$name(ledger).await;
            }
        };
    }

    #[macro_export]
    macro_rules! storage_test_suite {
        () => {
            $crate::storage_unit_test!(transaction);
            $crate::storage_unit_test!(transaction_does_not_update_stale_revision);
            $crate::storage_unit_test!(transaction_not_available_until_commit);
            $crate::storage_unit_test!(payments_always_include_negative_amounts);
            $crate::storage_unit_test!(does_not_update_spent_payments);
            $crate::storage_unit_test!(does_not_spend_unspendable_payments);
            $crate::storage_unit_test!(spend_spendable_payments);
            $crate::storage_unit_test!(relate_account_to_transaction);
            $crate::storage_unit_test!(find_transactions_by_tags);
            $crate::storage_unit_test!(find_transactions_by_status);
            $crate::storage_unit_test!(not_spendable_new_payments_not_spendable);
            $crate::storage_unit_test!(subscribe_realtime);
        };
    }

    pub async fn transaction_does_not_update_stale_revision<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let config = Config {
            storage,
            status: StatusManager::default(),
        };

        let ledger = Ledger::new(config);

        let asset: Asset = "USD/2".parse().expect("valid asset");
        let deposit = Transaction::new_external_deposit(
            "test reference".to_owned(),
            "pending".into(),
            vec![],
            vec![(
                "alice".parse().expect("account"),
                asset.from_human("100.99").expect("valid amount"),
            )],
        )
        .expect("valid tx");

        let deposit = ledger.store(deposit).await.expect("valid insert");

        ledger
            .change_status(
                deposit.revision_id.clone(),
                "processing".into(),
                "some text".to_owned(),
            )
            .await
            .expect("valid updated");

        ledger
            .change_status(
                deposit.revision_id.clone(),
                "processing".into(),
                "some text".to_owned(),
            )
            .await
            .expect_err("stale updates are rejected by storage");
    }

    pub async fn transaction<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let asset: Asset = "USD/2".parse().expect("valid asset");
        let deposit = Transaction::new_external_deposit(
            "test reference".to_owned(),
            "settled".into(),
            vec![],
            vec![(
                "alice".parse().expect("account"),
                asset.from_human("100.99").expect("valid amount"),
            )],
        )
        .expect("valid tx");
        let mut storing = storage.begin().await.expect("valid tx");
        storing
            .store_base_transaction(&deposit.id, &deposit.transaction)
            .await
            .expect("store tx");
        storing
            .store_revision(&deposit.revision_id, &deposit.revision)
            .await
            .expect("store revision");
        storing
            .update_transaction_revision(
                &deposit.id,
                &deposit.revision_id,
                deposit.revision.previous.as_ref(),
            )
            .await
            .expect("update tx");
        storing.rollback().await.expect("rollback");
        assert!(storage
            .find(deposit.id.clone().into())
            .await
            .expect("result")
            .is_empty());

        let mut storing = storage.begin().await.expect("valid tx");
        storing
            .store_base_transaction(&deposit.id, &deposit.transaction)
            .await
            .expect("store tx");
        storing
            .store_revision(&deposit.revision_id, &deposit.revision)
            .await
            .expect("store revision");
        storing
            .update_transaction_revision(
                &deposit.id,
                &deposit.revision_id,
                deposit.revision.previous.as_ref(),
            )
            .await
            .expect("update tx");
        storing.commit().await.expect("commit");
        assert_eq!(
            storage.find(deposit.id.into()).await.expect("result").len(),
            1
        );
    }

    pub async fn transaction_not_available_until_commit<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let usd: Asset = "USD/2".parse().expect("valid asset");
        let deposit = Transaction::new_external_deposit(
            "test reference".to_owned(),
            "settled".into(),
            vec![],
            vec![(
                "alice".parse().expect("account"),
                usd.from_human("100.99").expect("valid amount"),
            )],
        )
        .expect("valid tx");
        let mut storing = storage.begin().await.expect("valid tx");
        storing
            .store_base_transaction(&deposit.id, &deposit.transaction)
            .await
            .expect("store tx");
        storing
            .store_revision(&deposit.revision_id, &deposit.revision)
            .await
            .expect("store revision");
        storing
            .update_transaction_revision(
                &deposit.id,
                &deposit.revision_id,
                deposit.revision.previous.as_ref(),
            )
            .await
            .expect("update tx");
        assert!(storage
            .find(deposit.id.clone().into())
            .await
            .expect("results")
            .is_empty());
        storing.rollback().await.expect("rollback");
        assert!(storage
            .find(deposit.id.clone().into())
            .await
            .expect("results")
            .is_empty());

        let mut storing = storage.begin().await.expect("valid tx");
        storing
            .store_base_transaction(&deposit.id, &deposit.transaction)
            .await
            .expect("store base tx");
        storing
            .store_revision(&deposit.revision_id, &deposit.revision)
            .await
            .expect("store revision");
        storing
            .update_transaction_revision(
                &deposit.id,
                &deposit.revision_id,
                deposit.revision.previous.as_ref(),
            )
            .await
            .expect("update tx");
        storing.commit().await.expect("commit");
        assert_eq!(
            1,
            storage
                .find(deposit.id.into())
                .await
                .expect("results")
                .len()
        );
    }

    pub async fn does_not_update_spent_payments<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let mut writer = storage.begin().await.expect("writer");
        let mut rng = rand::thread_rng();
        let account = "account0".parse::<AccountId>().expect("account");
        let usd: Asset = "USD/2".parse().expect("valid asset");
        let amount = usd
            .from_human(&format!("{}", rng.gen_range(-1000.0..1000.0)))
            .expect("valid amount");

        let transaction_id = vec![0u8; 32].try_into().expect("valid tx id");

        let recipient = PaymentTo {
            to: account.clone(),
            amount,
        };

        let recipients = vec![recipient];

        writer
            .create_payments(&transaction_id, &recipients, ReceivedPaymentStatus::Locked)
            .await
            .expect("valid payment");

        // Alter state
        assert_eq!(
            writer
                .update_transaction_payments(
                    &transaction_id,
                    ReceivedPaymentStatus::Locked,
                    ReceivedPaymentStatus::Locked
                )
                .await
                .expect("valid"),
            (1, 0)
        );

        // Transactions is settled and their spent payments are forever spent
        assert_eq!(
            writer
                .update_transaction_payments(
                    &transaction_id,
                    ReceivedPaymentStatus::Spent,
                    ReceivedPaymentStatus::Spent
                )
                .await
                .expect("valid"),
            (1, 0)
        );

        for status in [
            ReceivedPaymentStatus::Failed,
            ReceivedPaymentStatus::Spendable,
        ] {
            assert_eq!(
                writer
                    .update_transaction_payments(&transaction_id, status, status)
                    .await
                    .expect("valid update"),
                (0, 0),
                "status: {:?}",
                status
            );
        }
    }

    pub async fn subscribe_realtime<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let ledger = Ledger::new(Config {
            storage,
            status: Default::default(),
        });

        let (_, mut subscription) = ledger
            .subscribe(Filter {
                tags: vec!["even".parse().expect("valid tag")],
                ..Default::default()
            })
            .await;

        assert_eq!(1, ledger.subscribers().await);

        for i in 0..10 {
            let usd: Asset = "USD/2".parse().expect("valid asset");
            let account = format!("account-{}", i).parse().expect("valid account");

            let deposit = ledger
                .deposit(
                    &account,
                    usd.from_human(&format!("10{}.99", i))
                        .expect("valid amount"),
                    "settled".into(),
                    vec![],
                    format!("test deposit {}", i),
                )
                .await
                .expect("valid deposit");

            if i % 2 == 0 {
                ledger
                    .set_tags(
                        deposit.revision_id,
                        vec![
                            "even".parse().expect("valid tag"),
                            "all".parse().expect("valid tag"),
                        ],
                        "add tags".to_owned(),
                    )
                    .await
                    .expect("tag tx");
            } else {
                ledger
                    .set_tags(
                        deposit.revision_id,
                        vec![
                            "odd".parse().expect("valid tag"),
                            "all".parse().expect("valid tag"),
                        ],
                        "add tags".to_owned(),
                    )
                    .await
                    .expect("tag tx");
            }
        }

        sleep(Duration::from_secs(1)).await;

        let expectactions = vec!["100.99", "102.99", "104.99", "106.99", "108.99", ""];

        assert_eq!(1, ledger.subscribers().await);
        for expectation in expectactions {
            assert_eq!(
                expectation.to_string(),
                subscription
                    .try_recv()
                    .map(|t| t.creates[0].amount.to_string())
                    .unwrap_or_default()
            )
        }
        assert_eq!(1, ledger.subscribers().await);

        drop(subscription);

        // TODO: Update this test when the drop() triggers the subscribers hashmap to cleanup
        assert_eq!(1, ledger.subscribers().await);

        let usd: Asset = "USD/2".parse().expect("valid asset");
        let account = "account-99".parse().expect("valid account");

        ledger
            .deposit(
                &account,
                usd.from_human(&"1010.99").expect("valid amount"),
                "settled".into(),
                vec!["even".parse().expect("valid tag")],
                "test deposit after the subscription listener is dropped".to_owned(),
            )
            .await
            .expect("valid deposit");

        sleep(Duration::from_secs(1)).await;

        assert_eq!(0, ledger.subscribers().await);

        drop(ledger);
        sleep(Duration::from_secs(1)).await;
    }

    pub async fn find_transactions_by_tags<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let ledger = Ledger::new(Config {
            storage,
            status: Default::default(),
        });

        for i in 0..10 {
            let usd: Asset = "USD/2".parse().expect("valid asset");
            let account = format!("account-{}", i).parse().expect("valid account");

            let deposit = ledger
                .deposit(
                    &account,
                    usd.from_human("100.99").expect("valid amount"),
                    "settled".into(),
                    vec![],
                    format!("test deposit {}", i),
                )
                .await
                .expect("valid deposit");

            if i % 2 == 0 {
                ledger
                    .set_tags(
                        deposit.revision_id,
                        vec![
                            "even".parse().expect("valid tag"),
                            "even".parse().expect("valid tag"),
                            "all".parse().expect("valid tag"),
                        ],
                        "add tags".to_owned(),
                    )
                    .await
                    .expect("tag tx");
            } else {
                ledger
                    .set_tags(
                        deposit.revision_id,
                        vec![
                            "odd".parse().expect("valid tag"),
                            "all".parse().expect("valid tag"),
                        ],
                        "add tags".to_owned(),
                    )
                    .await
                    .expect("tag tx");
            }
        }

        assert_eq!(
            5,
            ledger
                .get_transactions(Filter {
                    tags: vec!["even".parse().expect("valid tag")],
                    ..Default::default()
                })
                .await
                .expect("valid filter")
                .len()
        );

        assert_eq!(
            5,
            ledger
                .get_transactions(Filter {
                    tags: vec!["odd".parse().expect("valid tag")],
                    ..Default::default()
                })
                .await
                .expect("valid filter")
                .len()
        );

        assert_eq!(
            10,
            ledger
                .get_transactions(Filter {
                    tags: vec!["all".parse().expect("valid tag")],
                    ..Default::default()
                })
                .await
                .expect("valid filter")
                .len()
        );
    }

    pub async fn find_transactions_by_status<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let ledger = Ledger::new(Config {
            storage,
            status: Default::default(),
        });

        for i in 0..10 {
            let usd: Asset = "USD/2".parse().expect("valid asset");
            let account = format!("account-{}", i).parse().expect("valid account");

            ledger
                .deposit(
                    &account,
                    usd.from_human("100.99").expect("valid amount"),
                    if i % 2 == 0 {
                        "even".into()
                    } else {
                        "odd".into()
                    },
                    vec![],
                    format!("test deposit {}", i),
                )
                .await
                .expect("valid deposit");
        }

        assert_eq!(
            5,
            ledger
                .get_transactions(Filter {
                    status: vec!["even".parse().expect("valid tag")],
                    ..Default::default()
                })
                .await
                .expect("valid filter")
                .len()
        );

        assert_eq!(
            5,
            ledger
                .get_transactions(Filter {
                    status: vec!["odd".parse().expect("valid tag")],
                    ..Default::default()
                })
                .await
                .expect("valid filter")
                .len()
        );

        assert_eq!(
            10,
            ledger
                .get_transactions(Filter {
                    status: vec![
                        "even".parse().expect("valid tag"),
                        "odd".parse().expect("valid tag")
                    ],
                    ..Default::default()
                })
                .await
                .expect("valid filter")
                .len()
        );
    }

    pub async fn spend_spendable_payments<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let mut writer = storage.begin().await.expect("writer");
        let mut rng = rand::thread_rng();
        let account = "account0".parse::<AccountId>().expect("account");
        let usd: Asset = "USD/2".parse().expect("valid asset");
        let amount = usd
            .from_human(&format!("{}", rng.gen_range(-1000.0..1000.0)))
            .expect("valid amount");

        for (i, status) in [
            ReceivedPaymentStatus::Locked,
            ReceivedPaymentStatus::Spendable,
        ]
        .into_iter()
        .enumerate()
        {
            let transaction_id: TxId = vec![i as u8; 32].try_into().expect("valid tx id");

            writer
                .create_payments(
                    &transaction_id,
                    &[PaymentTo {
                        to: account.clone(),
                        amount: amount.clone(),
                    }],
                    status,
                )
                .await
                .expect("valid payment");

            for (spend_new_status, create_new_status) in
                [(ReceivedPaymentStatus::Spent, ReceivedPaymentStatus::Spent)]
            {
                assert_eq!(
                    writer
                        .update_transaction_payments(
                            &transaction_id,
                            spend_new_status,
                            create_new_status
                        )
                        .await
                        .expect("no updates"),
                    (1, 0)
                );
            }
        }
    }

    pub async fn does_not_spend_unspendable_payments<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let mut writer = storage.begin().await.expect("writer");
        let mut rng = rand::thread_rng();
        let account = "account0".parse::<AccountId>().expect("account");
        let usd: Asset = "USD/2".parse().expect("valid asset");
        let amount = usd
            .from_human(&format!("{}", rng.gen_range(-1000.0..1000.0)))
            .expect("valid amount");

        for (i, status) in [ReceivedPaymentStatus::Failed, ReceivedPaymentStatus::Spent]
            .into_iter()
            .enumerate()
        {
            let transaction_id: TxId = vec![i as u8; 32].try_into().expect("valid tx id");

            writer
                .create_payments(
                    &transaction_id,
                    &[PaymentTo {
                        to: account.clone(),
                        amount: amount.clone(),
                    }],
                    status,
                )
                .await
                .expect("valid payment");

            for (spend_new_status, create_new_status) in [
                (ReceivedPaymentStatus::Spent, ReceivedPaymentStatus::Spent),
                (
                    ReceivedPaymentStatus::Spendable,
                    ReceivedPaymentStatus::Spendable,
                ),
            ] {
                assert_eq!(
                    writer
                        .update_transaction_payments(
                            &transaction_id,
                            spend_new_status,
                            create_new_status
                        )
                        .await
                        .expect("no updates"),
                    (0, 0)
                );
            }
        }
    }

    pub async fn payments_always_include_negative_amounts<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let mut writer = storage.begin().await.expect("writer");
        let accounts: Vec<AccountId> = (0..10)
            .map(|i| format!("account{}", i).parse().expect("account"))
            .collect();

        let mut rng = rand::thread_rng();
        let target_inputs_per_account = 20;
        let usd: Asset = "USD/2".parse().expect("valid asset");

        let mut negative_payments_per_account = HashMap::new();

        for (index, account) in accounts.iter().enumerate() {
            let transaction_id: TxId = vec![index as u8; 32].try_into().expect("valid tx id");
            let mut negative_payments: usize = 0;
            let recipients = (0..target_inputs_per_account)
                .map(|_| {
                    let amount = usd
                        .from_human(&format!("{}", rng.gen_range(-1000.0..1000.0)))
                        .expect("valid amount");

                    if amount.cents().is_negative() {
                        negative_payments += 1;
                    }

                    PaymentTo {
                        to: account.clone(),
                        amount,
                    }
                })
                .collect::<Vec<_>>();

            negative_payments_per_account.insert(account.clone(), negative_payments);

            writer
                .create_payments(
                    &transaction_id,
                    &recipients,
                    ReceivedPaymentStatus::Spendable,
                )
                .await
                .expect("valid payment");
        }
        writer.commit().await.expect("valid commit");

        let mut at_least_one_negative_amount = false;
        for account in &accounts {
            let balance = &storage.get_balance(account).await.expect("valid balance")[0];
            if balance.cents() < 0 {
                continue;
            }
            let all_unspent_amounts = storage
                .get_unspent_payments(account, balance.asset(), balance.cents())
                .await
                .expect("valid unspent payments")
                .into_iter()
                .map(|x| x.amount.cents())
                .collect::<Vec<_>>();

            let expected_negative_payments = all_unspent_amounts
                .iter()
                .filter(|x| x.is_negative())
                .count();

            assert_eq!(
                Some(expected_negative_payments),
                negative_payments_per_account.get(account).cloned()
            );
            assert_eq!(target_inputs_per_account, all_unspent_amounts.len());

            if !at_least_one_negative_amount {
                at_least_one_negative_amount = expected_negative_payments > 0;
            }
        }
        assert!(at_least_one_negative_amount);
    }

    pub async fn relate_account_to_transaction<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let account1: AccountId = "alice1".parse().expect("account");
        let account2: AccountId = "alice2".parse().expect("account");
        let account3: AccountId = "alice3".parse().expect("account");
        let usd: Asset = "USD/2".parse().expect("USD");
        let deposit = Transaction::new_external_deposit(
            "test reference".to_owned(),
            "settled".into(),
            vec![],
            vec![(
                account1.clone(),
                usd.from_human("100.99").expect("valid amount"),
            )],
        )
        .expect("valid tx");

        let mut batch = storage.begin().await.expect("valid tx");
        batch
            .store_base_transaction(&deposit.id, &deposit.transaction)
            .await
            .expect("is ok");
        batch
            .store_revision(&deposit.revision_id, &deposit.revision)
            .await
            .expect("store revision");
        batch
            .update_transaction_revision(
                &deposit.id,
                &deposit.revision_id,
                deposit.revision.previous.as_ref(),
            )
            .await
            .expect("update tx");
        batch
            .relate_account_to_transaction(&deposit.id, &account1, Type::Deposit)
            .await
            .expect("relationship");
        batch
            .relate_account_to_transaction(&deposit.id, &account2, Type::Deposit)
            .await
            .expect("relationship");

        batch.commit().await.expect("valid commit");

        let filter = Filter {
            typ: vec![Type::Deposit],
            ..Default::default()
        };

        assert_eq!(
            1,
            storage
                .find(filter.clone().account(account1.clone()))
                .await
                .expect("valid tx")
                .len()
        );
        assert_eq!(
            1,
            storage
                .find(filter.clone().account(account2.clone()))
                .await
                .expect("valid tx")
                .len()
        );
        assert_eq!(
            0,
            storage
                .find(filter.account(account3.clone()))
                .await
                .expect("valid tx")
                .len()
        );
        for account in [account1, account2, account3] {
            assert_eq!(
                0,
                storage
                    .find(Filter {
                        typ: vec![Type::Withdrawal],
                        accounts: vec![account],
                        ..Default::default()
                    })
                    .await
                    .expect("valid tx")
                    .len()
            );
        }
    }

    pub async fn not_spendable_new_payments_not_spendable<T>(storage: T)
    where
        T: Storage + Send + Sync,
    {
        let mut writer = storage.begin().await.expect("writer");
        let mut rng = rand::thread_rng();
        let account = "account0".parse::<AccountId>().expect("account");
        let usd: Asset = "USD/2".parse().expect("valid asset");
        let amount = usd
            .from_human(&format!("{}", rng.gen_range(10.0..10000.0)))
            .expect("valid amount");

        let transaction: TxId = vec![0u8; 32].try_into().expect("valid tx id");

        writer
            .create_payments(
                &transaction,
                &[PaymentTo {
                    to: account.clone(),
                    amount,
                }],
                ReceivedPaymentStatus::Locked,
            )
            .await
            .expect("valid payment");

        writer.commit().await.expect("is ok");

        assert!(storage
            .get_balance(&account)
            .await
            .expect("valid balance")
            .is_empty());
    }
}
