use crate::{
    payment::PaymentTo,
    storage::{self, to_bytes, Error, ReceivedPaymentStatus},
    AccountId, BaseTx, PaymentId, RevId, Revision, Tag, TxId, Type,
};
use sqlx::{Row, Sqlite, Transaction as SqlxTransaction};
use std::{marker::PhantomData, num::TryFromIntError};

/// Creates a new Batch for SQLite
///
/// Batches are a group of updates to the databases, in which all of the are
/// executed or none. Same concept as a rdbms transaction.
pub struct Batch<'a> {
    inner: SqlxTransaction<'a, Sqlite>,
    x: PhantomData<&'a ()>,
}

impl<'a> Batch<'a> {
    /// Creates a new instance
    pub fn new(inner: SqlxTransaction<'a, Sqlite>) -> Batch<'a> {
        Self {
            inner,
            x: PhantomData,
        }
    }
}

#[async_trait::async_trait]
impl<'a> storage::Batch<'a> for Batch<'a> {
    async fn rollback(self) -> Result<(), Error> {
        self.inner
            .rollback()
            .await
            .map_err(|e| Error::Storage(e.to_string()))
    }

    async fn relate_account_to_transaction(
        &mut self,
        transaction: &TxId,
        account: &AccountId,
        typ: Type,
    ) -> Result<(), Error> {
        sqlx::query(
            r#"
            INSERT INTO "transaction_accounts" ("transaction_id", "account_id", "type")
            VALUES (?, ?, ?)
            "#,
        )
        .bind(transaction.to_string())
        .bind(account.to_string())
        .bind::<u32>(typ.into())
        .execute(&mut *self.inner)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;
        Ok(())
    }

    async fn commit(self) -> Result<(), Error> {
        self.inner
            .commit()
            .await
            .map_err(|e| Error::Storage(e.to_string()))
    }

    async fn spend_payments(
        &mut self,
        transaction_id: &TxId,
        spends: Vec<PaymentId>,
        status: ReceivedPaymentStatus,
    ) -> Result<(), Error> {
        let spend_ids = spends.iter().map(|id| id.to_string()).collect::<Vec<_>>();
        let placeholder = format!("?{}", ", ?".repeat(spend_ids.len()));

        let sql = format!(
            r#"
            UPDATE "payments"
            SET "status" = ?, "spent_by" = ?
            WHERE "payment_id" IN ({placeholder})
            AND "spent_by" IS NULL
        "#
        );

        let mut query = sqlx::query(&sql)
            .bind::<u32>(status.into())
            .bind(transaction_id.to_string());

        for id in spend_ids {
            query = query.bind(id);
        }

        let updated_records = usize::try_from(
            query
                .execute(&mut *self.inner)
                .await
                .map_err(|e| Error::Storage(e.to_string()))?
                .rows_affected(),
        )
        .map_err(|_| Error::Overflow)?;

        if updated_records == spends.len() {
            Ok(())
        } else {
            Err(Error::NoUpdate)
        }
    }

    async fn create_payments(
        &mut self,
        transaction_id: &TxId,
        recipients: &[PaymentTo],
        status: ReceivedPaymentStatus,
    ) -> Result<(), Error> {
        for (pos, recipient) in recipients.iter().enumerate() {
            let cents = recipient.amount.cents();
            sqlx::query(
                r#"
                INSERT INTO "payments"("payment_id", "to", "status", "asset", "cents", "is_negative")
                VALUES(?, ?, ?, ?, ?, ?)
            "#,
            )
            .bind(
                PaymentId {
                    transaction: transaction_id.clone(),
                    position: pos
                        .try_into()
                        .map_err(|e: TryFromIntError| Error::Encoding(e.to_string()))?,
                }
                .to_string(),
            )
            .bind(recipient.to.to_string())
            .bind::<u32>(status.into())
            .bind(recipient.amount.asset().to_string())
            .bind(cents.to_string())
            .bind(if cents.is_negative() { 1 } else { 0 })
            .execute(&mut *self.inner)
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;
        }

        Ok(())
    }

    async fn update_transaction_payments(
        &mut self,
        transaction_id: &TxId,
        create_status: ReceivedPaymentStatus,
        spend_status: ReceivedPaymentStatus,
    ) -> Result<(usize, usize), Error> {
        let creates = sqlx::query(
            r#"
            UPDATE "payments"
            SET "status" = ?, updated_at = datetime('now')
            WHERE "payment_id" LIKE ? AND "status" IN (?, ?)
        "#,
        )
        .bind::<u32>(create_status.into())
        .bind(format!("{}:%", transaction_id))
        .bind::<u32>(ReceivedPaymentStatus::Spendable.into())
        .bind::<u32>(ReceivedPaymentStatus::Locked.into())
        .execute(&mut *self.inner)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;

        let spends = sqlx::query(
            r#"
            UPDATE "payments"
            SET "status" = ?, "spent_by" = ?, updated_at = datetime('now')
            WHERE "spent_by" = ? AND "status" IN (?, ?)
            "#,
        )
        .bind::<u32>(spend_status.into())
        .bind(if spend_status == ReceivedPaymentStatus::Spendable {
            None
        } else {
            Some(transaction_id.to_string())
        })
        .bind(transaction_id.to_string())
        .bind::<u32>(ReceivedPaymentStatus::Spendable.into())
        .bind::<u32>(ReceivedPaymentStatus::Locked.into())
        .execute(&mut *self.inner)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;

        Ok((
            usize::try_from(creates.rows_affected()).map_err(|_| Error::Overflow)?,
            usize::try_from(spends.rows_affected()).map_err(|_| Error::Overflow)?,
        ))
    }

    async fn get_payment_status(
        &mut self,
        transaction_id: &TxId,
    ) -> Result<Option<ReceivedPaymentStatus>, Error> {
        let row = sqlx::query(
            r#"
            SELECT
                "p"."status"
            FROM
                "payments" "p"
            WHERE
                "p"."payment_id" = ?
            LIMIT 1
            "#,
        )
        .bind(transaction_id.to_string())
        .fetch_optional(&mut *self.inner)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;

        if let Some(row) = row {
            let status = row
                .try_get::<u32, usize>(0)
                .map_err(|_| Error::Storage("failed to parse status".to_owned()))?;

            status
                .try_into()
                .map(Some)
                .map_err(|_| Error::Storage("failed to parse status".to_owned()))
        } else {
            return Ok(None);
        }
    }

    async fn store_base_transaction(
        &mut self,
        transaction_id: &TxId,
        transaction: &BaseTx,
    ) -> Result<(), Error> {
        sqlx::query(
            r#"
                INSERT INTO "base_transactions"("transaction_id", "type", "blob")
                VALUES(?, ?, ?)
            "#,
        )
        .bind(transaction_id.to_string())
        .bind::<u32>(transaction.typ.into())
        .bind(to_bytes(&transaction)?)
        .execute(&mut *self.inner)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;

        Ok(())
    }

    async fn update_transaction_revision(
        &mut self,
        transaction_id: &TxId,
        revision_id: &RevId,
        last_rev_id: Option<&RevId>,
    ) -> Result<(), Error> {
        let r = if let Some(last_rev_id) = last_rev_id {
            sqlx::query(
                r#"
            UPDATE "transactions"
            SET "previous_revision_id" = ?, "revision_id" = ?
            WHERE "transaction_id" = ? AND "revision_id" = ?
            "#,
            )
            .bind(last_rev_id.to_string())
            .bind(revision_id.to_string())
            .bind(transaction_id.to_string())
            .bind(last_rev_id.to_string())
        } else {
            sqlx::query(
                r#"
            INSERT INTO "transactions"("transaction_id", "revision_id", "status")
            VALUES(?, ?, "")
                "#,
            )
            .bind(transaction_id.to_string())
            .bind(revision_id.to_string())
        }
        .execute(&mut *self.inner)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;
        if r.rows_affected() == 0 {
            Err(Error::NoUpdate)
        } else {
            Ok(())
        }
    }

    async fn store_revision(
        &mut self,
        revision_id: &RevId,
        revision: &Revision,
    ) -> Result<(), Error> {
        sqlx::query(
            r#"
            INSERT INTO "revisions"("revision_id", "blob", "transaction_id", "previous_revision_id", "status")
            VALUES(?, ?, ?, ?, ?)
        "#,
        )
        .bind(revision_id.to_string())
        .bind(to_bytes(revision)?)
        .bind(revision.transaction_id().to_string())
        .bind(
            revision.previous_revision_id().map(|prev_rev| prev_rev.to_string())
        )
        .bind(revision.status.to_string())
        .execute(&mut *self.inner)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;

        Ok(())
    }

    async fn tag_transaction(
        &mut self,
        transaction_id: &TxId,
        base_tx: &BaseTx,
        tags: &[Tag],
    ) -> Result<(), Error> {
        sqlx::query(r#"DELETE FROM "transactions_by_tags" WHERE "transaction_id" = ? "#)
            .bind(transaction_id.to_string())
            .execute(&mut *self.inner)
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;

        for tag in tags {
            sqlx::query(r#"INSERT INTO "transactions_by_tags"("transaction_id", "tag", "created_at") VALUES(?, ?, ?)
                ON CONFLICT("transaction_id", "tag", "created_at") DO NOTHING
                "#)
                .bind(transaction_id.to_string())
                .bind(tag.as_str())
                .bind(base_tx.created_at.timestamp())
                .execute(&mut *self.inner)
                .await
                .map_err(|e| Error::Storage(e.to_string()))?;
        }

        Ok(())
    }
}
