//! SQLite storage layer for Verax
use super::{Cursor, ReceivedPaymentStatus};
use crate::{
    amount::AmountCents,
    storage::{Error, Storage},
    transaction::Revision,
    AccountId, Amount, Asset, BaseTx, Filter, PaymentFrom, PaymentId, PrimaryFilter, RevId, TxId,
};
use borsh::from_slice;
use futures::TryStreamExt;
use sqlx::{sqlite::SqliteRow, Executor, Row};
use std::{collections::HashMap, str::FromStr};

mod batch;

pub use batch::Batch;
pub use sqlx::sqlite::{SqliteConnectOptions, SqliteJournalMode, SqlitePoolOptions};

/// SQLite storage layer for Verax
pub struct SQLite {
    db: sqlx::SqlitePool,
}

/// Receives a cursor and does the initial loading of transactions that may match the requested
/// filter.
///
/// The function will load the full transaction of candidates and return them.
#[allow(warnings)]
async fn find_candidates<'e, 'c: 'e, E>(
    executor: E,
    cursor: &mut Cursor,
    limit: usize,
) -> Result<Vec<(BaseTx, Revision)>, Error>
where
    E: sqlx::Executor<'c, Database = sqlx::Sqlite>,
{
    let since = if let Some(since) = cursor.filter.since {
        format!("AND created_at <= {}", since.timestamp())
    } else {
        "".to_owned()
    };
    let until = if let Some(until) = cursor.filter.until {
        format!("AND created_at >= {}", until.timestamp())
    } else {
        "".to_owned()
    };

    let rows = match &cursor.primary_filter {
        PrimaryFilter::TxId(ids) => {
            let sql = format!(
                r#"
                    SELECT
                        "bt"."blob",
                        "b"."blob"
                    FROM
                        "transactions" as "t",
                        "base_transactions" as "bt",
                        "revisions" as "b"
                    WHERE
                        "t"."transaction_id" IN ({})
                        AND "t"."revision_id" = "b"."revision_id"
                        AND "t"."transaction_id" = "bt"."transaction_id"
                        {} {}
                    ORDER BY "t"."created_at" DESC
                    LIMIT {} OFFSET {}
                    "#,
                "?,".repeat(ids.len()).trim_end_matches(","),
                since,
                until,
                limit,
                cursor.filter.skip
            );
            let mut query = sqlx::query(&sql);
            for id in ids.iter() {
                query = query.bind(id.to_string());
            }
            query
                .bind(limit as i64)
                .bind(cursor.filter.skip as i64)
                .fetch_all(executor)
                .await
        }
        PrimaryFilter::Revision(rev_ids) => {
            let sql = format!(
                r#"
                    SELECT
                        "bt"."blob",
                        "b"."blob"
                    FROM
                        "base_transactions" as "bt",
                        "revisions" as "b"
                    WHERE
                        "b"."revision_id" IN ({})
                        AND "b"."transaction_id" = "bt"."transaction_id"
                        {} {}
                    ORDER BY "b"."created_at" DESC
                    LIMIT {} OFFSET {}
                    "#,
                "?,".repeat(rev_ids.len()).trim_end_matches(","),
                since,
                until,
                limit,
                cursor.filter.skip
            );
            let mut query = sqlx::query(&sql);
            for id in rev_ids.iter() {
                query = query.bind(id.to_string());
            }
            query.fetch_all(executor).await
        }
        PrimaryFilter::Account(account_ids) => {
            let sql = format!(
                r#"
                    SELECT
                        "bt"."blob",
                        "b"."blob"
                    FROM
                        "transaction_accounts" as "ta",
                        "transactions" as "t",
                        "base_transactions" as "bt",
                        "revisions" as "b"
                    WHERE
                        "ta"."account_id" IN ({})
                        AND "t"."transaction_id" = "ta"."transaction_id"
                        AND "t"."revision_id" = "b"."revision_id"
                        AND "t"."transaction_id" = "bt"."transaction_id"
                        {} {}
                    ORDER BY "ta"."created_at" DESC
                    LIMIT {} OFFSET {}
                    "#,
                "?,".repeat(account_ids.len()).trim_end_matches(","),
                since,
                until,
                limit,
                cursor.filter.skip
            );
            let mut query = sqlx::query(&sql);
            for id in account_ids.iter() {
                query = query.bind(id.to_string());
            }
            query.fetch_all(executor).await
        }
        PrimaryFilter::Type(types) => {
            let sql = format!(
                r#"
                    SELECT
                        "bt"."blob",
                        "b"."blob"
                    FROM
                        "transactions" as "t",
                        "base_transactions" as "bt",
                        "revisions" as "b"
                    WHERE
                        "bt"."type" IN ({})
                        AND "t"."revision_id" = "b"."revision_id"
                        AND "t"."transaction_id" = "bt"."transaction_id"
                        {} {}
                    ORDER BY "bt"."created_at" DESC
                    LIMIT {} OFFSET {}
                    "#,
                "?,".repeat(types.len()).trim_end_matches(","),
                since,
                until,
                limit,
                cursor.filter.skip
            );
            let mut query = sqlx::query(&sql);
            for id in types.iter() {
                query = query.bind(id.to_string());
            }
            query.fetch_all(executor).await
        }
        PrimaryFilter::Tags(tags) => {
            let sql = format!(
                r#"
                    SELECT
                        "bt"."blob",
                        "b"."blob"
                    FROM
                        "transactions" as "t",
                        "base_transactions" as "bt",
                        "revisions" as "b"
                    WHERE
                        "t"."transaction_id" IN (SELECT DISTINCT "transaction_id" FROM "transactions_by_tags" WHERE "tag" IN ({}) ORDER BY "created_at" DESC)
                        AND "t"."revision_id" = "b"."revision_id"
                        AND "t"."transaction_id" = "bt"."transaction_id"
                        {} {}
                    ORDER BY "t"."created_at" DESC
                    LIMIT {} OFFSET {}
                    "#,
                "?,".repeat(tags.len()).trim_end_matches(","),
                since,
                until,
                limit,
                cursor.filter.skip
            );

            let mut query = sqlx::query(&sql);
            for id in tags.iter() {
                query = query.bind(id.to_string());
            }
            query.fetch_all(executor).await
        }
        PrimaryFilter::Status(statuses) => {
            let sql = format!(
                r#"
                  SELECT
                      "bt"."blob",
                      "b"."blob"
                  FROM
                      "transactions" as "t",
                      "base_transactions" as "bt",
                      "revisions" as "b"
                  WHERE
                      "b"."status" IN ({})
                      AND "t"."revision_id" = "b"."revision_id"
                      AND "t"."transaction_id" = "bt"."transaction_id"
                      {} {}
                  ORDER BY "t"."created_at" DESC
                  LIMIT {} OFFSET {}
                  "#,
                "?,".repeat(statuses.len()).trim_end_matches(","),
                since,
                until,
                limit,
                cursor.filter.skip
            );
            let mut query = sqlx::query(&sql);
            for status in statuses.iter() {
                query = query.bind(status.to_string());
            }
            query.fetch_all(executor).await
        }
        PrimaryFilter::Stream => {
            let sql = format!(
                r#"
                    SELECT
                        "bt"."blob",
                        "b"."blob"
                    FROM
                        "transactions" as "t",
                        "base_transactions" as "bt",
                        "revisions" as "b"
                    WHERE
                        "t"."revision_id" = "b"."revision_id"
                        AND "t"."transaction_id" = "bt"."transaction_id"
                        {} {}
                    ORDER BY "t"."created_at" DESC
                    LIMIT {} OFFSET {}
                    "#,
                since, until, limit, cursor.filter.skip
            );
            sqlx::query(&sql).fetch_all(executor).await
        }
    };

    let results = rows
        .map_err(|e| Error::Storage(e.to_string()))?
        .into_iter()
        .map(|row| {
            let base_tx = row
                .try_get::<Vec<u8>, usize>(0)
                .map_err(|e| Error::Storage(e.to_string()))?;

            let revision = row
                .try_get::<Vec<u8>, usize>(1)
                .map_err(|e| Error::Storage(e.to_string()))?;

            Ok((
                from_slice::<BaseTx>(&base_tx)?,
                from_slice::<Revision>(&revision)?,
            ))
        })
        .collect::<Result<Vec<_>, Error>>()?;

    cursor.filter.skip = cursor
        .filter
        .skip
        .checked_add(results.len())
        .ok_or(Error::Overflow)?;

    Ok(results)
}

impl SQLite {
    /// Creates a new Verax SQLite storage layer. It takes an instance of the asset_manager
    pub fn new(db: sqlx::SqlitePool) -> Self {
        Self { db }
    }

    /// Creates all the tables and indexes that are needed
    /// The database uses string for the primary keys to make things easier to debug[1]
    ///
    /// [1] https://sqlite.org/forum/info/69c8cbbe2d84410f
    pub async fn setup(&self) -> Result<(), sqlx::Error> {
        let mut x = self.db.begin().await?;
        x.execute(
            r#"
        CREATE TABLE IF NOT EXISTS "transactions" (
            "transaction_id" VARCHAR(67) NOT NULL PRIMARY KEY,
            "previous_revision_id" VARCHAR(68) DEFAULT NULL,
            "revision_id" VARCHAR(68) DEFAULT NULL,
            "status" VARCHAR(30) NOT NULL,
            "created_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
            "updated_at" DATETIME DEFAULT CURRENT_TIMESTAMP
        );
        CREATE TABLE IF NOT EXISTS "base_transactions" (
            "transaction_id" VARCHAR(67) NOT NULL PRIMARY KEY,
            "type" INTEGER NOT NULL,
            "blob" BINARY NOT NULL,
            "created_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
            "updated_at" DATETIME DEFAULT CURRENT_TIMESTAMP
        );
        CREATE TABLE IF NOT EXISTS "revisions" (
            "revision_id" VARCHAR(68) NOT NULL PRIMARY KEY,
            "transaction_id" VARCHAR(67) NOT NULL,
            "previous_revision_id" VARCHAR(68) DEFAULT NULL,
            "status" VARCHAR(30) NOT NULL,
            "blob" BINARY NOT NULL,
            "created_at" DATETIME DEFAULT CURRENT_TIMESTAMP
        );
        CREATE TABLE IF NOT EXISTS "transactions_by_tags" (
            "transaction_id" VARCHAR(67),
            "tag" VARCHAR(250) NOT NULL,
            "created_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
            PRIMARY KEY ("tag", "created_at", "transaction_id")
        );
        CREATE INDEX IF NOT EXISTS "transaction_id_and_tags" ON "transactions_by_tags" ("transaction_id");
        CREATE TABLE IF NOT EXISTS "payments" (
            "payment_id" VARCHAR(80) NOT NULL PRIMARY KEY,
            "to" VARCHAR(64) NOT NULL,
            "status" INT NOT NULL,
            "asset" VARCHAR(10) NOT NULL,
            "cents" BIGINT NOT NULL,
            "is_negative" INT DEFAULT '0',
            "spent_by" VARCHAR(64) DEFAULT NULL,
            "created_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
            "updated_at" DATETIME DEFAULT CURRENT_TIMESTAMP
        );
        CREATE INDEX IF NOT EXISTS "spent_by" ON "payments" ("to", "status", "spent_by", "is_negative");
        CREATE TABLE IF NOT EXISTS "transaction_accounts" (
            "id" INTEGER PRIMARY KEY AUTOINCREMENT,
            "account_id" VARCHAR(64) NOT NULL,
            "transaction_id" VARCHAR(66) NOT NULL,
            "type" INTEGER NOT NULL,
            "created_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
            "updated_at" DATETIME DEFAULT CURRENT_TIMESTAMP
        );
        CREATE UNIQUE INDEX IF NOT EXISTS "unique_account_transaction" ON "transaction_accounts" ("account_id", "transaction_id");
        CREATE INDEX IF NOT EXISTS "sorted_account_transaction" ON "transaction_accounts" ("account_id", "id" desc);
        "#,
        )
        .await?;
        x.commit().await?;
        Ok(())
    }

    fn sql_row_to_amount(row: &SqliteRow, pos: usize) -> Result<Amount, Error> {
        let asset: Asset = row
            .try_get::<String, usize>(pos)
            .map_err(|_| Error::Storage("Invalid asset".to_string()))?
            .parse()
            .map_err(|_| Error::Storage("Invalid asset".to_string()))?;

        let cents = row
            .try_get::<i64, usize>(pos.checked_add(1).ok_or(Error::Overflow)?)
            .map_err(|_| Error::Storage("Invalid cents".to_string()))?
            .into();

        Ok(asset.new_amount(cents))
    }

    fn sql_row_to_payment_id(row: &SqliteRow, pos: usize) -> Result<PaymentId, Error> {
        row.try_get::<String, usize>(pos)
            .map_err(|_| Error::Storage("Invalid payment_id".to_string()))?
            .as_str()
            .parse()
            .map_err(|_| Error::Storage("Invalid payment_id length".to_string()))
    }
}

#[async_trait::async_trait]
impl Storage for SQLite {
    type Batch<'a> = Batch<'a>
    where
        Self: 'a;

    async fn begin<'a>(&'a self) -> Result<Self::Batch<'a>, Error> {
        self.db
            .begin()
            .await
            .map(Batch::new)
            .map_err(|x| Error::Storage(x.to_string()))
    }

    async fn get_balance(&self, account: &AccountId) -> Result<Vec<Amount>, Error> {
        let mut conn = self
            .db
            .acquire()
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;

        let mut result = sqlx::query(
            r#"
            SELECT
                "asset",
                SUM("cents") as "cents"
            FROM
                "payments"
            WHERE
                "to" = ?
                AND status = ?
                AND "spent_by" IS NULL
            GROUP BY "asset"
            "#,
        )
        .bind(account.to_string())
        .bind::<u32>(ReceivedPaymentStatus::Spendable.into())
        .fetch(&mut *conn);

        let mut balances = HashMap::<Asset, Amount>::new();

        while let Some(row) = result
            .try_next()
            .await
            .map_err(|e| Error::Storage(e.to_string()))?
        {
            let asset: Asset = row
                .try_get::<String, usize>(0)
                .map_err(|_| Error::Storage("Invalid asset_id".to_string()))?
                .parse()
                .map_err(|_| Error::Storage("Invalid asset_id".to_string()))?;

            let cents = row
                .try_get::<i64, usize>(1)
                .map_err(|_| Error::Storage("Invalid cents".to_string()))?
                .into();

            let new_amount = asset.new_amount(cents);

            if let Some(amount) = balances.get_mut(&asset) {
                *amount = amount
                    .checked_add(&new_amount)
                    .ok_or(Error::Storage("amount overflow".to_owned()))?;
            } else {
                balances.insert(asset, new_amount);
            }
        }

        Ok(balances.into_values().collect())
    }

    async fn get_negative_unspent_payments(
        &self,
        account: &AccountId,
        asset: &Asset,
    ) -> Result<Vec<PaymentFrom>, Error> {
        let mut conn = self
            .db
            .acquire()
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;
        sqlx::query(
            r#"
            SELECT
                "p"."payment_id",
                "p"."asset",
                "p"."cents",
                "p"."to"
            FROM
                "payments" as "p"
            WHERE
                "p"."to" = ?
                AND "p"."asset" = ?
                AND "p"."status" = ?
                AND "p"."spent_by" IS NULL
                AND "p"."is_negative" = 1
            "#,
        )
        .bind(account.to_string())
        .bind(asset.to_string())
        .bind::<u32>(ReceivedPaymentStatus::Spendable.into())
        .fetch_all(&mut *conn)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?
        .into_iter()
        .map(|row| {
            let amount = Self::sql_row_to_amount(&row, 1)?;
            Ok(PaymentFrom {
                id: Self::sql_row_to_payment_id(&row, 0)?,
                from: row
                    .try_get::<String, usize>(3)
                    .map_err(|_| Error::Storage("Invalid from".to_string()))?
                    .parse()
                    .map_err(|_| Error::Encoding("Invalid account encoding".to_owned()))?,
                amount,
            })
        })
        .collect::<Result<Vec<_>, Error>>()
    }

    async fn get_positive_unspent_payments(
        &self,
        account: &AccountId,
        asset: &Asset,
        mut target_amount: AmountCents,
    ) -> Result<Vec<PaymentFrom>, Error> {
        let mut conn = self
            .db
            .acquire()
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;
        let results = sqlx::query(
            r#"
            SELECT
                "p"."payment_id",
                "p"."asset",
                "p"."cents",
                "p"."to"
            FROM
                "payments" as "p"
            WHERE
                "p"."to" = ?
                AND "p"."asset" = ?
                AND "p"."status" = ?
                AND "p"."spent_by" IS NULL
                AND "p"."is_negative" = 0
            ORDER BY cents ASC
            "#,
        )
        .bind(account.to_string())
        .bind(asset.to_string())
        .bind::<u32>(ReceivedPaymentStatus::Spendable.into())
        .fetch_all(&mut *conn)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;

        let mut to_return = vec![];
        for row in results.into_iter() {
            let amount = Self::sql_row_to_amount(&row, 1)?;
            target_amount = target_amount
                .checked_sub(amount.cents())
                .ok_or(Error::Underflow)?;
            to_return.push(PaymentFrom {
                id: Self::sql_row_to_payment_id(&row, 0)?,
                from: row
                    .try_get::<String, usize>(3)
                    .map_err(|_| Error::Storage("Invalid from".to_string()))?
                    .parse()
                    .map_err(|_| Error::Encoding("Invalid account encoding".to_owned()))?,
                amount,
            });
            if target_amount <= 0 {
                break;
            }
        }

        if target_amount <= 0 {
            Ok(to_return)
        } else {
            Err(Error::NotEnoughUnspentPayments(target_amount))
        }
    }

    async fn get_revisions(&self, transaction_id: &TxId) -> Result<Vec<RevId>, Error> {
        let mut conn = self
            .db
            .acquire()
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;

        Ok(sqlx::query(
            r#"
            SELECT
                "revision_id"
            FROM
                "revisions"
            WHERE
                "transaction_id" = ?
            ORDER BY "created_at" ASC
            "#,
        )
        .bind(transaction_id.to_string())
        .fetch_all(&mut *conn)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?
        .into_iter()
        .map(|x| {
            x.try_get::<String, usize>(0)
                .map(|x| RevId::from_str(&x))
                .map_err(|e| Error::Storage(e.to_string()))
        })
        .collect::<Result<Result<Vec<_>, _>, _>>()??)
    }

    async fn find_base_tx_and_revision(
        &self,
        filter: Filter,
    ) -> Result<Vec<(BaseTx, Revision)>, Error> {
        let mut conn = self
            .db
            .acquire()
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;

        let mut cursor: Cursor = filter.into();

        let candidate_limits = 20;
        let mut results = vec![];

        loop {
            let candidates = find_candidates(&mut *conn, &mut cursor, candidate_limits).await?;
            // if there are fewer candidates than requested it means there are no more resultset
            let no_more_candidates = candidates.len() != candidate_limits;

            let mut iteration_result = candidates
                .into_iter()
                .filter(|(base, revision)| cursor.matches(base, revision))
                .collect::<Vec<_>>();

            results.append(&mut iteration_result);

            if no_more_candidates
                || (cursor.filter.limit != 0 && cursor.filter.limit >= results.len())
            {
                break;
            }
        }

        if cursor.filter.limit != 0 {
            results.truncate(cursor.filter.limit);
        }
        Ok(results)
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::storage_test_suite;
    use sqlx::sqlite::{SqliteConnectOptions, SqlitePoolOptions};
    use std::fs::remove_file;

    async fn get_instance(test: &str) -> SQLite {
        let path = format!("/tmp/{}.db", test);
        let _ = remove_file(&path);
        let settings = path
            .parse::<SqliteConnectOptions>()
            .expect("valid settings")
            .journal_mode(sqlx::sqlite::SqliteJournalMode::Wal)
            .create_if_missing(true);

        let pool = SqlitePoolOptions::new()
            .connect_with(settings)
            .await
            .expect("pool");

        let db = SQLite::new(pool);
        db.setup().await.expect("setup");
        db
    }

    storage_test_suite!();
}
