use super::{deposit, get_asset_manager_and_ledger};
use crate::{AccountId, Asset};

#[tokio::test]
async fn negative_deposit_prevent_spending() {
    let source = "account1".parse::<AccountId>().expect("account");
    let dest = "account2".parse::<AccountId>().expect("account");
    let fee = "fee".parse::<AccountId>().expect("account");
    let ledger = get_asset_manager_and_ledger().await;
    let usd: Asset = "USD/2".parse().expect("asset");

    // Deposit some money
    deposit(&ledger, &source, usd.from_human("50").expect("amount")).await;
    // Take money of source's account
    deposit(&ledger, &source, usd.from_human("-100").expect("amount")).await;

    assert_eq!(
        vec![usd.from_human("-50").expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );

    // Try to spend money
    assert_eq!(
        "Storage: Not enough unspent payments (missing 6000 cents)".to_owned(),
        ledger
            .new_transaction(
                "Exchange one".to_owned(),
                "settled".into(),
                vec![(source.clone(), usd.from_human("10").expect("amount"))],
                vec![
                    (dest.clone(), usd.from_human("9.5").expect("amount")),
                    (fee.clone(), usd.from_human("0.5").expect("amount")),
                ],
            )
            .await
            .unwrap_err()
            .to_string()
    );
}

#[tokio::test]
async fn negative_deposit_prevent_spending_payback() {
    let source = "account1".parse::<AccountId>().expect("account");
    let dest = "account2".parse::<AccountId>().expect("account");
    let fee = "fee".parse::<AccountId>().expect("account");
    let ledger = get_asset_manager_and_ledger().await;
    let usd: Asset = "USD/2".parse().expect("asset");

    // Deposit some money
    deposit(&ledger, &source, usd.from_human("50").expect("amount")).await;
    // Take money of source's account
    deposit(&ledger, &source, usd.from_human("-100").expect("amount")).await;

    assert_eq!(
        vec![usd.from_human("-50").expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );

    // Try to spend money
    assert_eq!(
        "Storage: Not enough unspent payments (missing 6000 cents)".to_owned(),
        ledger
            .new_transaction(
                "Exchange one".to_owned(),
                "settled".into(),
                vec![(source.clone(), usd.from_human("10").expect("amount"))],
                vec![
                    (dest.clone(), usd.from_human("9.5").expect("amount")),
                    (fee.clone(), usd.from_human("0.5").expect("amount")),
                ],
            )
            .await
            .unwrap_err()
            .to_string()
    );

    // Payback the debt
    deposit(&ledger, &source, usd.from_human("150").expect("amount")).await;

    ledger
        .new_transaction(
            "Exchange one".to_owned(),
            "settled".into(),
            vec![(source.clone(), usd.from_human("10").expect("amount"))],
            vec![
                (dest.clone(), usd.from_human("9.5").expect("amount")),
                (fee.clone(), usd.from_human("0.5").expect("amount")),
            ],
        )
        .await
        .expect("valid tx");

    assert_eq!(
        vec![usd.from_human("9.5").expect("amount")],
        ledger.get_balance(&dest).await.expect("balance")
    );
    assert_eq!(
        vec![usd.from_human("0.5").expect("amount")],
        ledger.get_balance(&fee).await.expect("balance")
    );
    assert_eq!(
        vec![usd.from_human("90").expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );
}
