use crate::{
    amount::AmountCents,
    payment::PaymentTo,
    storage::to_bytes,
    transaction::{Error, Revision, Type},
    AccountId, Asset, PaymentFrom, Status, TxId,
};
use chrono::{serde::ts_milliseconds, DateTime, Utc};
use serde::{Deserialize, Serialize};
use sha2::{Digest, Sha256};
use std::collections::HashMap;

#[derive(Debug, Clone, Deserialize, Serialize, borsh::BorshSerialize, borsh::BorshDeserialize)]
/// Transaction struct
///
/// A transaction is an immutable object that represents a financial operation. It contains a list
/// of payments that are spent and a list of payments that are created. The sum of the payments
/// spent must be equal to the sum of the payments created, unless the transaction is a Deposit or
/// Withdrawal transaction.
pub struct BaseTx {
    /// List of spend payments to create this transaction
    #[serde(skip_serializing_if = "Vec::is_empty")]
    pub spends: Vec<PaymentFrom>,
    /// New payments that are created by this transaction
    #[serde(skip_serializing_if = "Vec::is_empty")]
    pub creates: Vec<PaymentTo>,
    /// Transaction type
    #[serde(rename = "type")]
    pub typ: Type,
    /// Human-readable reference to this transaction
    pub reference: String,
    #[serde(rename = "created_at", with = "ts_milliseconds")]
    #[borsh(
        serialize_with = "super::to_ts_microseconds",
        deserialize_with = "super::from_ts_microseconds"
    )]
    /// Timestamp when this transaction has been created
    pub created_at: DateTime<Utc>,
}

/// Transaction object
impl BaseTx {
    /// Creates a new instance
    pub fn new(
        spends: Vec<PaymentFrom>,
        creates: Vec<PaymentTo>,
        reference: String,
        typ: Type,
        status: Status,
    ) -> Result<(Self, Revision), Error> {
        let new = Self {
            spends,
            creates,
            reference,
            typ,
            created_at: Utc::now(),
        };
        let id = new.id()?;
        Ok((new, Revision::new(id, None, "".to_owned(), vec![], status)))
    }

    /// Calculates the transaction ID
    pub fn id(&self) -> Result<TxId, Error> {
        let mut hasher = Sha256::new();
        hasher.update(to_bytes(&self)?);
        Ok(TxId::new(hasher.finalize().into()))
    }

    /// Validates the transaction input and output (debit and credit)
    ///
    /// The total sum of debits and credits should always be zero in transactions, unless they are
    /// deposits or withdrawals.
    ///
    /// Negative amounts can be used in transactions, but the total sum of debits and credits should
    /// always be zero, and the debit amount should be positive numbers
    pub fn validate(&self) -> Result<(), Error> {
        let mut debit = HashMap::<Asset, AmountCents>::new();
        let mut credit = HashMap::<Asset, AmountCents>::new();

        for input in self.spends.iter() {
            if let Some(value) = debit.get_mut(input.amount.asset()) {
                *value = input
                    .amount
                    .cents()
                    .checked_add(*value)
                    .ok_or(Error::Overflow)?;
            } else {
                debit.insert(input.amount.asset().clone(), input.amount.cents());
            }
        }

        for (asset, amount) in debit.iter() {
            if *amount <= 0 {
                return Err(Error::InvalidAmount(
                    asset.new_amount(*amount),
                    asset.new_amount(*amount),
                ));
            }
        }

        if !self.typ.is_transaction() {
            // We don't care input/output balance in external operations
            // (withdrawals/deposits), because these operations are inbalanced
            return Ok(());
        }

        for output in self.creates.iter() {
            if let Some(value) = credit.get_mut(output.amount.asset()) {
                *value = output
                    .amount
                    .cents()
                    .checked_add(*value)
                    .ok_or(Error::Overflow)?;
            } else {
                credit.insert(output.amount.asset().clone(), output.amount.cents());
            }
        }

        for (asset, credit_amount) in credit.into_iter() {
            if let Some(debit_amount) = debit.remove(&asset) {
                if debit_amount != credit_amount {
                    return Err(Error::InvalidAmount(
                        asset.new_amount(debit_amount),
                        asset.new_amount(credit_amount),
                    ));
                }
            } else {
                return Err(Error::MissingSpendingAsset(asset));
            }
        }

        if let Some((asset, _)) = debit.into_iter().next() {
            return Err(Error::MissingPaymentAsset(asset));
        }

        Ok(())
    }

    /// Returns a unique list of accounts involved in this transaction.
    ///
    /// Accounts are sorted and unique, and they include the accounts that spent and that receives
    pub fn accounts(&self) -> Vec<AccountId> {
        let mut accounts = self
            .creates
            .iter()
            .map(|x| x.to.clone())
            .collect::<Vec<_>>();

        accounts.extend(
            self.spends
                .iter()
                .map(|x| x.from.clone())
                .collect::<Vec<_>>(),
        );

        accounts.sort();
        accounts.dedup();
        accounts
    }
}
